@@ -9,11 +9,12 @@
     :license: BSD, see LICENSE for more details.
 """
 import json
-from threading import Thread
+import threading
+from threading import Thread, Lock
 import platform
 from logbook.base import NOTSET, LogRecord, dispatch_record
 from logbook.handlers import Handler, WrapperHandler
-from logbook.helpers import PY2
+from logbook.helpers import PY2, u
 
 if PY2:
     from Queue import Empty, Queue as ThreadQueue
@@ -21,20 +22,118 @@ else:
     from queue import Empty, Queue as ThreadQueue
 
 
-class RabbitMQHandler(Handler):
-    """A handler that acts as a RabbitMQ publisher, which publishes each record
-    as json dump.  Requires the kombu module.
+class RedisHandler(Handler):
+    """A handler that sends log messages to a Redis instance.
 
-    The queue will be filled with JSON exported log records.  To receive such
-    log records from a queue you can use the :class:`RabbitMQSubscriber`.
+    It publishes each record as json dump. Requires redis module.
 
+    To receive such records you need to have a running instance of Redis.
 
     Example setup::
 
-        handler = RabbitMQHandler('amqp://guest:guest@localhost//', queue='my_log')
+        handler = RedisHandler('http://127.0.0.1', port='9200', key='redis')
+
+    If your Redis instance is password protected, you can securely connect passing
+    your password when creating a RedisHandler object.
+
+    Example::
+
+        handler = RedisHandler(password='your_redis_password')
+
+    More info about the default buffer size: wp.me/p3tYJu-3b
     """
+    def __init__(self, host='127.0.0.1', port=6379, key='redis', extra_fields={},
+                flush_threshold=128, flush_time=1, level=NOTSET, filter=None,
+                password=False, bubble=True, context=None):
+        Handler.__init__(self, level, filter, bubble)
+        try:
+            import redis
+            from redis import ResponseError
+        except ImportError:
+            raise RuntimeError('The redis library is required for '
+                               'the RedisHandler')
+
+        self.redis = redis.Redis(host=host, port=port, password=password, decode_responses=True)
+        try:
+            self.redis.ping()
+        except ResponseError:
+            raise ResponseError('The password provided is apparently incorrect')
+        self.key = key
+        self.extra_fields = extra_fields
+        self.flush_threshold = flush_threshold
+        self.queue = []
+        self.lock = Lock()
+
+        #Set up a thread that flushes the queue every specified seconds
+        self._stop_event = threading.Event()
+        self._flushing_t = threading.Thread(target=self._flush_task,
+                                            args=(flush_time, self._stop_event))
+        self._flushing_t.daemon = True
+        self._flushing_t.start()
+
+
+    def _flush_task(self, time, stop_event):
+        """Calls the method _flush_buffer every certain time.
+        """
+        while not self._stop_event.isSet():
+            with self.lock:
+                self._flush_buffer()
+            self._stop_event.wait(time)
+
+
+    def _flush_buffer(self):
+        """Flushes the messaging queue into Redis.
+
+        All values are pushed at once for the same key.
+        """
+        if self.queue:
+            self.redis.rpush(self.key, *self.queue)
+        self.queue = []
+
+
+    def disable_buffering(self):
+        """Disables buffering.
+
+        If called, every single message will be directly pushed to Redis.
+        """
+        self._stop_event.set()
+        self.flush_threshold = 1
+
+
+    def emit(self, record):
+        """Emits a pair (key, value) to redis.
+
+        The key is the one provided when creating the handler, or redis if none was
+        provided. The value contains both the message and the hostname. Extra values
+        are also appended to the message.
+        """
+        with self.lock:
+            r = {"message": record.msg, "host": platform.node(), "level": record.level_name}
+            r.update(self.extra_fields)
+            r.update(record.kwargs)
+            self.queue.append(json.dumps(r))
+            if len(self.queue) == self.flush_threshold:
+                self._flush_buffer()
+
+
+    def close(self):
+        self._flush_buffer()
+
+
+class MessageQueueHandler(Handler):
+    """A handler that acts as a message queue publisher, which publishes each
+    record as json dump. Requires the kombu module.
+
+    The queue will be filled with JSON exported log records.  To receive such
+    log records from a queue you can use the :class:`MessageQueueSubscriber`.
+
+    Example setup::
+
+        handler = MessageQueueHandler('mongodb://localhost:27017/logging')
+    """
+
     def __init__(self, uri=None, queue='logging', level=NOTSET,
-                filter=None, bubble=False, context=None):
+                 filter=None, bubble=False, context=None):
         Handler.__init__(self, level, filter, bubble)
         try:
             import kombu
@@ -58,6 +157,9 @@ class RabbitMQHandler(Handler):
         self.queue.close()
 
 
+RabbitMQHandler = MessageQueueHandler
+
+
 class ZeroMQHandler(Handler):
     """A handler that acts as a ZeroMQ publisher, which publishes each record
     as json dump.  Requires the pyzmq library.
@@ -65,6 +167,11 @@ class ZeroMQHandler(Handler):
     The queue will be filled with JSON exported log records.  To receive such
     log records from a queue you can use the :class:`ZeroMQSubscriber`.
 
+    If `multi` is set to `True`, the handler will use a `PUSH` socket to
+    publish the records. This allows multiple handlers to use the same `uri`.
+    The records can be received by using the :class:`ZeroMQSubscriber` with
+    `multi` set to `True`.
+
 
     Example setup::
 
@@ -72,7 +179,7 @@ class ZeroMQHandler(Handler):
     """
 
     def __init__(self, uri=None, level=NOTSET, filter=None, bubble=False,
-                 context=None):
+                 context=None, multi=False):
         Handler.__init__(self, level, filter, bubble)
         try:
             import zmq
@@ -81,10 +188,18 @@ class ZeroMQHandler(Handler):
                                'the ZeroMQHandler.')
         #: the zero mq context
         self.context = context or zmq.Context()
-        #: the zero mq socket.
-        self.socket = self.context.socket(zmq.PUB)
-        if uri is not None:
-            self.socket.bind(uri)
+
+        if multi:
+            #: the zero mq socket.
+            self.socket = self.context.socket(zmq.PUSH)
+            if uri is not None:
+                self.socket.connect(uri)
+        else:
+            #: the zero mq socket.
+            self.socket = self.context.socket(zmq.PUB)
+            if uri is not None:
+                self.socket.bind(uri)
+
 
     def export_record(self, record):
         """Exports the record into a dictionary ready for JSON dumping."""
@@ -176,27 +291,27 @@ class SubscriberBase(object):
         return controller
 
 
-class RabbitMQSubscriber(SubscriberBase):
-    """A helper that acts as RabbitMQ subscriber and will dispatch received
-    log records to the active handler setup.  There are multiple ways to
-    use this class.
+class MessageQueueSubscriber(SubscriberBase):
+    """A helper that acts as a message queue subscriber and will dispatch
+    received log records to the active handler setup. There are multiple ways
+    to use this class.
 
     It can be used to receive log records from a queue::
 
-        subscriber = RabbitMQSubscriber('amqp://guest:guest@localhost//')
+        subscriber = MessageQueueSubscriber('mongodb://localhost:27017/logging')
         record = subscriber.recv()
 
     But it can also be used to receive and dispatch these in one go::
 
         with target_handler:
-            subscriber = RabbitMQSubscriber('amqp://guest:guest@localhost//')
+            subscriber = MessageQueueSubscriber('mongodb://localhost:27017/logging')
             subscriber.dispatch_forever()
 
     This will take all the log records from that queue and dispatch them
     over to `target_handler`.  If you want you can also do that in the
     background::
 
-        subscriber = RabbitMQSubscriber('amqp://guest:guest@localhost//')
+        subscriber = MessageQueueSubscriber('mongodb://localhost:27017/logging')
         controller = subscriber.dispatch_in_background(target_handler)
 
     The controller returned can be used to shut down the background
@@ -204,13 +319,11 @@ class RabbitMQSubscriber(SubscriberBase):
 
         controller.stop()
     """
-
     def __init__(self, uri=None, queue='logging'):
         try:
             import kombu
         except ImportError:
-            raise RuntimeError('The kombu library is required for '
-                               'the RabbitMQSubscriber.')
+            raise RuntimeError('The kombu library is required.')
         if uri:
             connection = kombu.Connection(uri)
 
@@ -245,6 +358,9 @@ class RabbitMQSubscriber(SubscriberBase):
         return LogRecord.from_dict(log_record)
 
 
+RabbitMQSubscriber = MessageQueueSubscriber
+
+
 class ZeroMQSubscriber(SubscriberBase):
     """A helper that acts as ZeroMQ subscriber and will dispatch received
     log records to the active handler setup.  There are multiple ways to
@@ -272,9 +388,14 @@ class ZeroMQSubscriber(SubscriberBase):
     thread::
 
         controller.stop()
+
+    If `multi` is set to `True`, the subscriber will use a `PULL` socket
+    and listen to records published by a `PUSH` socket (usually via a
+    :class:`ZeroMQHandler` with `multi` set to `True`). This allows a
+    single subscriber to dispatch multiple handlers.
     """
 
-    def __init__(self, uri=None, context=None):
+    def __init__(self, uri=None, context=None, multi=False):
         try:
             import zmq
         except ImportError:
@@ -284,11 +405,18 @@ class ZeroMQSubscriber(SubscriberBase):
 
         #: the zero mq context
         self.context = context or zmq.Context()
-        #: the zero mq socket.
-        self.socket = self.context.socket(zmq.SUB)
-        if uri is not None:
-            self.socket.connect(uri)
-        self.socket.setsockopt_unicode(zmq.SUBSCRIBE, u'')
+
+        if multi:
+            #: the zero mq socket.
+            self.socket = self.context.socket(zmq.PULL)
+            if uri is not None:
+                self.socket.bind(uri)
+        else:
+            #: the zero mq socket.
+            self.socket = self.context.socket(zmq.SUB)
+            if uri is not None:
+                self.socket.connect(uri)
+            self.socket.setsockopt_unicode(zmq.SUBSCRIBE, u(''))
 
     def __del__(self):
         try:
@@ -430,7 +558,7 @@ class ExecnetChannelSubscriber(SubscriberBase):
     def __init__(self, channel):
         self.channel = channel
 
-    def recv(self, timeout=-1):
+    def recv(self, timeout=None):
         try:
             rv = self.channel.receive(timeout=timeout)
         except self.channel.RemoteError:
@@ -540,7 +668,7 @@ class SubscriberGroup(SubscriberBase):
 
         subscribers = SubscriberGroup([
             MultiProcessingSubscriber(queue),
-            ZeroMQSubscriber('tcp://localhost:5000')
+            ZeroMQSubscriber('tcp://127.0.0.1:5000')
         ])
         with target_handler:
             subscribers.dispatch_forever()
